# Adapted from the Berkeley Pacman projects
# ---------------
# Licensing Information:  You are free to use or extend these projects for
# educational purposes provided that (1) you do not distribute or publish
# solutions, (2) you retain this notice, and (3) you provide clear
# attribution to UC Berkeley, including a link to http://ai.berkeley.edu.
# 
# Attribution Information: The Pacman AI projects were developed at UC Berkeley.
# The core projects and autograders were primarily created by John DeNero
# (denero@cs.berkeley.edu) and Dan Klein (klein@cs.berkeley.edu).
# Student side autograding was added by Brad Miller, Nick Hay, and
# Pieter Abbeel (pabbeel@cs.berkeley.edu).

####################
####################

import heapq
import inspect
import sys
from ast import literal_eval

####################
####################

# Convenience function to halt when
# a function is not defined
# (e.g. in a superclass)
# 
# See license above
def raiseNotDefined():
	fileName = inspect.stack()[1][1]
	line = inspect.stack()[1][2]
	method = inspect.stack()[1][3]

	print("*** Method not implemented: {} at line {} of {}".format(method, line, fileName))
	sys.exit(1)

####################
####################

# WARNING: setting to True will slow down all searches and explode memory
# (it maintains a set of all pushed nodes, for visualization purposes)
_STORE_NODES = False

# A LIFO stack!
# 
# See license above
class Stack:
	"A container with a last-in-first-out (LIFO) queuing policy."
	def __init__(self):
		self.list = []
		self.count = 0
		self.pushed = set()

	def push(self,item):
		"Push 'item' onto the stack"
		self.list.append(item)
		self.count += 1
		if _STORE_NODES:
			self.pushed.add(item)

	def pop(self):
		"Pop the most recently pushed item from the stack"
		return self.list.pop()

	def isEmpty(self):
		"Returns true if the stack is empty"
		return len(self.list) == 0

# A FIFO queue!
# 
# See license above
class Queue:
	"A container with a first-in-first-out (FIFO) queuing policy."
	def __init__(self):
		self.list = []
		self.count = 0
		self.pushed = set()

	def push(self,item):
		"Enqueue the 'item' into the queue"
		self.list.insert(0,item)
		self.count += 1
		if _STORE_NODES:
			self.pushed.add(item)

	def pop(self):
		"""
		Dequeue the earliest enqueued item still in the queue. This
		operation removes the item from the queue.
		"""
		return self.list.pop()

	def isEmpty(self):
		"Returns true if the queue is empty"
		return len(self.list) == 0

# A Priority Queue!
# 
# See license above
class PriorityQueue:
	"""
	  Implements a priority queue data structure. Each inserted item
	  has a priority associated with it and the client is usually interested
	  in quick retrieval of the lowest-priority item in the queue. This
	  data structure allows O(1) access to the lowest-priority item.

	  Note that this PriorityQueue does not allow you to change the priority
	  of an item.  However, you may insert the same item multiple times with
	  different priorities.
	"""
	def  __init__(self):
		self.heap = []
		self.count = 0
		self.pushed = set()

	def push(self, item, priority):
		entry = (priority, self.count, item)
		heapq.heappush(self.heap, entry)
		self.count += 1
		if _STORE_NODES:
			self.pushed.add(item)

	def pop(self):
		(_, _, item) = heapq.heappop(self.heap)
		return item

	def isEmpty(self):
		return len(self.heap) == 0

# A Priority Queue that allows
# you to pre-specify the cost
# function (thereby creating a
# common interface between
# Stack/Queue/PriorityQueue)
# 
# See license above
class PriorityQueueWithFunction(PriorityQueue):
	"""
	Implements a priority queue with the same push/pop signature of the
	Queue and the Stack classes. This is designed for drop-in replacement for
	those two classes. The caller has to provide a priority function, which
	extracts each item's priority.
	"""
	def  __init__(self, priorityFunction):
		"priorityFunction (item) -> priority"
		self.priorityFunction = priorityFunction      # store the priority function
		PriorityQueue.__init__(self)      	# super-class initializer

	def push(self, item):
		"Adds an item to the queue with priority from the priority function"
		PriorityQueue.push(self, item, self.priorityFunction(item))

####################
####################

# Valid directions
# 
# See license above
class Directions:
	NORTH = 'North'
	SOUTH = 'South'
	EAST = 'East'
	WEST = 'West'
	STOP = 'Stop'

# Directions that correspond
# to search actions
# 
# See license above
class Actions:
	"""
	A collection of static methods for manipulating move actions.
	"""
	_directions = { Directions.NORTH: (0, 1),
					Directions.SOUTH: (0, -1),
					Directions.EAST:  (1, 0),
					Directions.WEST:  (-1, 0),
					Directions.STOP:  (0, 0)}

	def directionToVector(direction, speed = 1.0):
		dx, dy =  Actions._directions[direction]
		return (dx * speed, dy * speed)
	directionToVector = staticmethod(directionToVector)

# Superclass for task-specific
# search problems
# 
# See license above
class SearchProblem:
	"""
	This class outlines the structure of a search problem, but doesn't implement
	any of the methods (in object-oriented terminology: an abstract class).

	You do not need to change anything in this class, ever.
	"""

	def getStartState(self):
		"""
		Returns the start state for the search problem.
		"""
		raiseNotDefined()

	def isGoalState(self, state):
		"""
		state: Search state

		Returns True if and only if the state is a valid goal state.
		"""
		raiseNotDefined()

	def getSuccessors(self, state):
		"""
		state: Search state

		For a given state, this should return a list of triples, (successor,
		action, stepCost), where 'successor' is a successor to the current
		state, 'action' is the action required to get there, and 'stepCost' is
		the incremental cost of expanding to that successor.
		"""
		raiseNotDefined()

	def getCostOfActions(self, actions):
		"""
		actions: A list of actions to take

		This method returns the total cost of a particular sequence of actions.
		The sequence must be composed of legal moves.
		"""
		raiseNotDefined()

	def whereNext(self, state, action):
		nextstates = self.getSuccessors(state)
		for ns in nextstates:
			if ns[1] == action:
				return (ns[0], ns[2])

####################
####################

# Input: path to a maze file
# Output: (startx, starty),(finishx, finishy),[[maze row 1],[maze row 2],...]
def readMaze(fpath):
	with open(fpath, 'rU') as f:
		start = literal_eval(f.readline().strip())
		finish = literal_eval(f.readline().strip())
		
		maze = []
		for line in f:
			maze.append(list(line.strip()))
		
		return start, finish, maze
